.First.lib <- function(lib, pkg)
{
   cat("ISIS for R>=1.4.0, Anja von Heydebreck\n")
   library.dynam("isis", pkg, lib)
   library(mva)
   
}

#------------------------------------------------------------
# R-wrapper around the C-function "ttesttwo"
#------------------------------------------------------------
ttesttwo <- function(dat, split) {
  if ((!is.vector(split)) || (length(split)!=ncol(dat)))
    stop(paste("Fatal error in ttesttwo: second argument \"split\" is not a vector, or length(split)=",
               length(split), " but ncol(dat)=",
               ncol(dat), " - they should be the same.\n", sep=""))
  if(any(is.na(dat)))
     stop("Fatal error in ttesttwo: first argument \"dat\" contains NA values.\n")
     
  res <- .C("isis", "ttesttwo", 
            as.double(t(dat)),                  
            as.integer(nrow(dat)),
            as.integer(ncol(dat)),
            as.integer(split),
            as.integer(1),
            as.integer(c(0,0,0)),
            tstat      = numeric(nrow(dat)), 
            returncode = as.integer(0),       
            DUP = TRUE)
  
  if (res$returncode) stop(paste("Fatal error in ttesttwo: caught exception", res$returncode))
  return(res$tstat)
}

#------------------------------------------------------------
# R-wrapper around the C-function "tscore"
#------------------------------------------------------------
tscore <- function(dat, split, p = 50, p.offs = 0) {
  if (is.vector(split)) { ncs <- length(split); nrs <- 1          }  
  if (is.matrix(split)) { ncs <- ncol(split);   nrs <- nrow(split)}
  if (ncs != ncol(dat))
    stop(paste("Fatal error in tscore: number of columns of second argument \"split\" is", ncs,
               "but first argument \"dat\" has", ncol(dat), "columns.",
               "They should be the same.\n"))
  if(any(is.na(dat)))
     stop("Fatal error in tscore: first argument \"dat\" contains NA values.\n")
  
  means <- matrix(rep(     apply(dat, 1, mean), ncs), nrow=nrow(dat), ncol=ncs)
  stds  <- matrix(rep(sqrt(apply(dat, 1, var)), ncs), nrow=nrow(dat), ncol=ncs)
    
  res <- .C("isis", "tscore", 
            as.double(t((dat-means)/stds)),       
            as.integer(nrow(dat)),
            as.integer(ncol(dat)),
            as.integer(split),    
            as.integer(nrs),
            as.integer(c(p, p.offs, 0)),
            t = numeric(nrs),
            returncode = as.integer(0),       
            DUP = TRUE)
  
  if (res$returncode) stop(paste("Fatal error in tscore: caught exception", res$returncode))
  return (res$t)
}

#------------------------------------------------------------
# R-wrapper around the C-function "gotomax"
#------------------------------------------------------------
gotomax <- function(dat, split, p = 50, p.offs = 0, min.split.size = ceiling(0.1*ncol(dat))) {
  if (is.vector(split)) { ncs <- length(split); nrs <- 1          }  
  if (is.matrix(split)) { ncs <- ncol(split);   nrs <- nrow(split)}
  if (ncs != ncol(dat))
    stop(paste("Fatal error in gotomax: number of columns of second argument \"split\" is", ncs,
               "but first argument \"dat\" has", ncol(dat), "columns.",
               "They should be the same.\n"))
  if(any(is.na(dat)))
     stop("Fatal error in gotomax: first argument \"dat\" contains NA values.\n")
  
  means <- matrix(rep(     apply(dat, 1, mean), ncs), nrow=nrow(dat), ncol=ncs)
  stds  <- matrix(rep(sqrt(apply(dat, 1, var)), ncs), nrow=nrow(dat), ncol=ncs)

  res <- .C("isis", "gotomax", 
            as.double(t((dat-means)/stds)),       
            as.integer(nrow(dat)),
            as.integer(ncol(dat)),
            spl = as.integer(split),    
            as.integer(nrs),
            as.integer(c(p, p.offs, min.split.size)),
            t = numeric(nrs), 
            returncode = as.integer(0),       
            DUP = TRUE)
  
  if (res$returncode!=0) stop(paste("Fatal error in gotomax: caught exception", res$returncode))
  
  gmres <- matrix(nrow = nrs, ncol = ncs+1)
  gmres[, 1:ncs  ] <- res$spl       # local maximum splits
  gmres[, (ncs+1)] <- res$t         # scores
  return (gmres)               
}

#------------------------------------------------------------
#------------------------------------------------------------
isis <- function (x, p = 50, p.offs = 0, max.nr.cand.splits = 500) {
  data(thresh)
  options(warn=1) # warnings are printed as they occur
  x <- as.matrix(x)
  nrgenes <- nrow(x)
  nrsample   <- ncol(x)
  if (nrsample > 150) stop("Fatal error in isis: too many samples; max. number = 150\n")
  if (nrsample < 10) stop("Fatal error in isis: not enough samples; min. number = 10\n")
  thresh <- as.vector(as.matrix(thresh[nrsample,1:(nrsample-1)]))

  # the minimum number of elements in each half of a bipartition
  min.split.size <- ceiling(0.1*nrsample)

  #--------------------------------------------------
  # enlarge data matrix with average profiles of clusters of genes
  #--------------------------------------------------
  cat("Clustering genes...\n")
  cx <- matrix(nrow = 2*nrgenes-1,ncol=nrsample)
  cx[1:nrgenes,] <- x 
  nrobjects <- rep(1, 2*nrgenes-1)
      
  d <- as.dist(1 - cor(t(x))) 
  c <- hclust(d, method = "centroid")
  # shift the indices of c$merge
  c$merge[which(c$merge>0)] <- c$merge[which(c$merge>0)] + nrgenes
  c$merge[which(c$merge<0)] <- -c$merge[which(c$merge<0)]
  for (j in 1:(nrgenes-1)) {
    cx[nrgenes + j,] <- nrobjects[c$merge[j,1]]*cx[c$merge[j,1],] + nrobjects[c$merge[j,2]]*cx[c$merge[j,2],]
    nrobjects[nrgenes + j] <- nrobjects[c$merge[j,1]] + nrobjects[c$merge[j,2]]
    cx[nrgenes + j,] <- (cx[nrgenes + j,] - mean(cx[nrgenes + j,]))/sqrt(var(cx[nrgenes + j,]))
  }
    
  cat("Sorting data matrix...\n")
  sox  <- t(apply(cx, 1, sort))   # sort row-wise
  ord  <- t(apply(cx, 1, order))  # record ordering

  #----------------------------------------------------------------------
  # Generating candidate partitions using the ordered rows of cx
  #----------------------------------------------------------------------
  cat("Generating candidate splits\n")
  candidate.splits <- new.env(hash=T)
  for (col in min.split.size : (nrsample - min.split.size)) {
    split <- c(rep(1, col), rep(0, nrsample-col))
    t <- ttesttwo(sox, split)
    for (k in which(t >= thresh[col])) {
      s <- rep(0, nrsample)
      s[ord[k,1:col]] <-1
      if (s[1]==0) s <- 1-s
      # cat(c('.','%')[s+1], '\n', sep="")
      key <- paste(c("a","b")[1+s], collapse="")
      assign(key, 0, envir=candidate.splits)
    }
  }

  key2vec <- function(key) {
    as.numeric(unlist(strsplit(key, split=""))=="b")
  }
  
  lsp <- lapply(ls(envir=candidate.splits), key2vec)
  candsplit <- matrix(unlist(lsp), nrow=length(lsp), ncol=nrsample, byrow=T)
                      
  # cat('\n'); for (j in 1:nrow(candsplit)) cat(c('.','%')[candsplit[j,]+1], '\n', sep="")
  
  #------------------------------------------------------------
  # Selecting the candidate partitions with highest score,
  # merging similar ones by clustering
  #------------------------------------------------------------
  cat("Computing scores of", nrow(candsplit), "candidate splits\n")
  weight <- tscore(x, candsplit, p=p, p.offs=p.offs)
  
  sorted.weights <- -sort(-weight)
  ordered <- order(-weight)
  
  result <- candsplit[ordered[1:min(max.nr.cand.splits, nrow(candsplit))],]
  
  cat("Clustering", nrow(result), "candidate splits")
  
  d1 <- dist(result, method="manhattan")
  d2 <- nrsample - d1
  d <- pmin(d1, d2)
  hc <- hclust(d, method = "complete")
  cluster <- cutree(hc, h=round(nrsample*0.1))
  clusteredsplits <- vector(mode="numeric", length=0)
  for (j in 1:max(cluster)) {
    temp <- matrix(result[which(cluster==j),], nrow=length(which(cluster==j)), ncol=nrsample)
    if (nrow(temp) > 1){
      for (k in 2:nrow(temp)) {
        if (length(which(temp[1,]==temp[k,])) < 0.5*nrsample) { # flip partition k
          temp[k,] <- 1-temp[k,]  
        }
      }
    }  
    temp <- sorted.weights[which(cluster==j)] %*% temp/sum(sorted.weights[which(cluster==j)])
    clusteredsplits <- c(clusteredsplits, temp)
  }
  clusteredsplits <- round(clusteredsplits)
  clusteredsplits <- t(matrix(clusteredsplits, nrow=nrsample, ncol=(length(clusteredsplits)/nrsample)))
  
  #------------------------------------------------------------
  # Improving partitions with local search, removing duplicate partitions
  #------------------------------------------------------------

  cat("\nPerforming local search with", nrow(clusteredsplits),  "candidates...\n")
  gmres <- gotomax(x, clusteredsplits, p=p, p.offs=p.offs, min.split.size=min.split.size)
    
  finres <- matrix(nrow=nrow(gmres), ncol=nrsample+1)
  finres[1,] <- gmres[1, 1:(nrsample+1)]
  k <- 1                      
  for (j in 2:nrow(gmres)) {
    if (!all(gmres[j,1:nrsample] == finres[k,1:nrsample]) &
        !all(gmres[j,1:nrsample] != finres[k,1:nrsample]) & gmres[j,nrsample+1] > 0)  {
      k <- k+1
      finres[k,] <- gmres[j, 1:(nrsample+1)]
    }
  }
  for (r in 1:k)
    cat(c(".","%")[finres[r, 1:nrsample]+1], " ", finres[r, nrsample+1], "\n", sep="")
  finres[1:k,]
}
